// code by jph

/*
PORTC (MUX0,MUX1,MUX2)
ADC0 odometry left/right?
ADC1 odometry left/right?
ADC2 front sensor left/right?
ADC3 front sensor left/right?
ADC4 switches
ADC5 battery
PC6 [reset]

PORTB
PB0 status led green
PB1 pwm
PB2 pwm
PB3 [supplies PD1]
PB4 motor direction left/right?
PB5 motor direction
PB6 [quarz]
PB7 [quarz]

PORTD
PD0 [uart rx]
PD1 [uart tx]
PD2 status led red
PD3 switches
PD4 motor direction left/right?
PD5 motor direction
PD6 front led
PD7 odometry
*/


#include "asuro.h"

//#define PROGRAM 
#define DELIM 0xff        // start byte of downlink/uplink sequence

unsigned char data[5];    // uart downlink/uplink sequence
unsigned int  valu[6];    // adc values

unsigned char prog[64];   // program commands with parameters
unsigned char speed=0x6E; // initial default speed

unsigned char eval=0;

// uart
unsigned char getBlock(void) {
	unsigned char len=0;
	unsigned char jmp=0;
	unsigned char tmp;
	unsigned int cnt=0;
	UCSRB=0x10;
	while (len<5&&cnt++<0x0a00)
		if (UCSRA&0x80) {
			tmp=UDR;
			if (tmp==DELIM&&jmp==0) {
				len=0;
				jmp=1;
			}
			data[len++]=tmp;
			cnt=0;
		}
	return data[0]==DELIM&&len==5&&(data[0]^data[1]^data[2]^data[3])==data[4];
}

void putBlock(void) {
  unsigned char len=0;
  data[0]=DELIM;
  data[4]=data[0]^data[1]^data[2]^data[3];  
  UCSRB=0x08;
	while (len<5)
		if (UCSRA&0x20)
			UDR=data[len++];
	while (!(UCSRA&0x40));
}

// AD conversion
void read_adc(unsigned char adc_input) {
	ADMUX=adc_input|1<<REFS0;	
	ADCSRA|=1<<ADSC;
	while ((ADCSRA&1<<ADIF)==0) ;
	ADCSRA|=1<<ADIF;
	data[2]=ADCH;
	data[3]=ADCL;
	valu[adc_input]=ADCW;
}

#ifdef PROGRAM

void run(void) {
	unsigned int c0,c1,c2;
	unsigned int refn;
	char h1=data[3];
	switch (data[2]) {
		case 0: //wait
			PORTD|=1<<PD6;
			for (c1=0;c1<(unsigned int)data[3]<<2;c1++)
				for (c2=0;c2<0xffff;c2++) ;
			PORTD^=1<<PD6;
			break;
		case 1: //speed
			speed=data[3];
			break;
		case 2: //drive
			OCR1A=OCR1B=0;	
			PORTD&=~(1<<PD4|1<<PD5);
			PORTB&=~(1<<PB4|1<<PB5);
			PORTD|=1<<PD4+(data[3]>127);
			PORTB|=1<<PB4+(data[3]>127);
			OCR1A=OCR1B=speed;
			for (c1=0;c1<(h1<0?-h1:h1);c1++)
				for (c2=0;c2<0xffff;c2++) ;
			OCR1A=OCR1B=0;
			break;
		case 3: //turn
			OCR1A=OCR1B=0;
			PORTD&=~(1<<PD4|1<<PD5);
			PORTB&=~(1<<PB4|1<<PB5);
			PORTD|=1<<PD4+(data[3]>127);
			PORTB|=1<<PB4+(data[3]<128);
			OCR1A=OCR1B=speed;
			for (c1=0;c1<(h1<0?-h1:h1);c1++)
				for (c2=0;c2<0xffff;c2++) ;
			OCR1A=OCR1B=0;
			break;
		case 4: //light
			PORTD|=1<<PD6;
			OCR1A=OCR1B=0;
			PORTD&=~(1<<PD4|1<<PD5);
			PORTB&=~(1<<PB4|1<<PB5);
			PORTD|=1<<PD4;
			PORTB|=1<<PB4;			
			getLightL();
			getLightR();
			refn=valu[0]+valu[1];
			refn*=4;
			refn/=5;
			OCR1A=OCR1B=speed;
			while (valu[0]+valu[1]>refn) {
				getLightL();
				getLightR();
			}
			PORTD^=1<<PD6;
			OCR1A=OCR1B=0;			
			break;
		case 5: //collide
			PORTD|=1<<PD3;
			for (c2=0;c2<0xffff;c2++) ;
			getContact();
			refn=valu[4];
			PORTD&=~(1<<PD4|1<<PD5); PORTD|=1<<PD4; //fwd direction
			PORTB&=~(1<<PB4|1<<PB5); PORTB|=1<<PB4; //fwd direction
			OCR1A=OCR1B=speed;			
			while (1) {
  			for (c2=0;c2<0x0fff;c2++) ;
				getContact();
				if (valu[4]<refn-100) break;
			}
			OCR1A=OCR1B=0;
			PORTD^=1<<PD3;
			break;
		case 6:	//side-move
			PORTD&=~(1<<PD4|1<<PD5);
			PORTB&=~(1<<PB4|1<<PB5);			
			OCR1A=OCR1B=speed;
			for (c0=0;c0<data[3];c0++) {
				PORTD^=1<<PD4;
				for (c1=0;c1<2;c1++)
					for (c2=0;c2<0xffff;c2++) ;
				PORTB^=1<<PB4;
				for (c1=0;c1<2;c1++)
					for (c2=0;c2<0xffff;c2++) ;
				PORTD^=1<<PD4;
				for (c1=0;c1<2;c1++)
					for (c2=0;c2<0xffff;c2++) ;					
				PORTB^=1<<PB4;
				PORTD^=1<<PD5;
				for (c1=0;c1<2;c1++)
					for (c2=0;c2<0xffff;c2++) ;
				PORTB^=1<<PB5;
				for (c1=0;c1<2;c1++)
					for (c2=0;c2<0xffff;c2++) ;
				PORTD^=1<<PD5;
				for (c1=0;c1<2;c1++)
					for (c2=0;c2<0xffff;c2++) ;
				PORTB^=1<<PB5;
			}
			OCR1A=OCR1B=0;			
			break;
  }
}

#endif

int main(void) {	
	unsigned int c1,c2;	
	unsigned char tmp;
	Init();
	OCR1A=0;
	OCR1B=0;
	sei();
	
	/* permanent communication to terminal */
	while (1) {
		if (getBlock()) {
			PORTB|=  1<<PB0 ; //status led green on
			PORTD&=~(1<<PD2); //status led red off
			switch (data[1]) {//comm
				case 32:
// UPLINK
// data[2].0 motor direction left
// data[2].1 motor direction right
// data[2].2 front led
// data[2].3 rear led left
// data[2].4 rear led right
					PORTC&=~(1<<PC0|1<<PC1);
					PORTC|=data[2]&(1<<PC0|1<<PC1);
					PORTD&=~(1<<PD6);
					PORTD|=data[2]&(1<<PD6); 
					tmp=1<<PD4; if (data[2]&0x10) tmp<<=1;
					PORTD=PORTD&~(1<<PD4|1<<PD5)|tmp;
					tmp=1<<PB4; if (data[2]&0x20) tmp<<=1;
					PORTB=PORTB&~(1<<PB4|1<<PB5)|tmp;
					break;
				case 33:
// UPLINK
// data[2] motor speed left
// data[3] motor speed right
					OCR1A=data[2];
					OCR1B=data[3];
					break;
				case 36:
// DOWNLINK
// data[2] motor speed left
// data[3] motor speed right
          putBlock();    //hack: send trash
					PORTD|=1<<PD3; //switches
					for (c1=2;c1<6;c1++)
					  data[1]=160+c1;
					  read_adc(c1);
					  putBlock();
				  }
					PORTD^=1<<PD3;					
					break;
				case 40: //demo
					putBlock();
					data[1]+=128;
					putBlock();
					run();
					putBlock();
					break;
				case 41: //prog
					putBlock();
					data[1]+=128;
					tmp=0;
					for (c1=0;c1<data[2]<<1;c1++)
						tmp^=prog[c1];
					if (tmp==data[3]) {
						eval=data[2];
						data[2]=254;
					}
					putBlock();
					break;
				case 42: //exec
					putBlock();
					data[1]+=128;
					putBlock();
					for (c1=0;c1<eval<<1;c1+=2) {
						data[2]=prog[c1+0];
						data[3]=prog[c1+1];
						run();
					}
					break;
			}
			if (data[1]&64) {// prog
				c1=(data[1]&63)<<1;
				prog[c1+0]=data[2];
				prog[c1+1]=data[3];
			}
		} else {
			PORTB&=~(1<<PB0); //status led green off
			PORTD|=  1<<PD2 ; //status led red on
		}
	}
	return 0;
}
				